@@ -1,4 +1,5 @@
 #!/usr/bin/env python
+from cgitb import reset
 from urllib.parse import parse_qsl
 import fire
 import os
@@ -27,8 +28,11 @@ def get_time():
 def get_host():
     return socket.gethostname().split('.', 1)[0] + " " + str(os.getpid())
 
+def time_diff(now, st):
+    return str(now - st).split('.')[0].replace(' day, ', '-')
+
 def table_list(data, headers = None):
-    return tabulate(data, headers = headers)
+    return tabulate(data, headers = headers, tablefmt="plain")
 
 def _log(info, color):
     csi = '\033['
@@ -77,7 +81,8 @@ class CAM(object):
                 pass
         if hasattr(self, "running_job_id"):
             try:
-                self._remove_by_tid("running", self.running_job_id)
+                self.kill(self.running_job_id)
+                self._store_finished_task()
                 os.killpg(os.getpgid(self.p.pid), signal.SIGINT)
                 #self.p.kill()
             except:
@@ -119,6 +124,17 @@ class CAM(object):
             return True
         else:
             return eval(cond)
+
+    def _store_finished_task(self):
+        data = self._get_by_tid("running", self.running_job_id)
+        st = datetime.datetime.fromisoformat(data[1])
+        now = datetime.datetime.utcnow()
+        peroid = time_diff(now, st)
+        data[1] = peroid
+        data[-1] = ("Finished " if not data[-1].startswith("KILLED") else "") + data[-1]
+        data.append(get_time())
+        self._set_by_tid("finished", self.running_job_id, json.dumps(data))
+        self._remove_by_tid("running", self.running_job_id)
         
 
     def server(self, port = None):
@@ -151,11 +167,11 @@ class CAM(object):
                 if not hasattr(self, "server_disconnected") or self.server_disconnected:
                     self.server_disconnected = False
                     log_info("Server Connected.")
-                    log_info("Server: ", self._conf['server'], ":", str(self._conf['port']), ' v', self.__version__)
-                if cnt <= 0:
+                    log_info(" ".join(["Server:", self._conf['server']+":"+str(self._conf['port']), ' v'+self.__version__, cond, cmdprefix, cmdsuffix]))
+                if not self._condition_parse(cond):
+                    self._redis.hset("workers", self.worker_id, json.dumps([worker_start_time, "Wait Resource", cond, cmdprefix, cmdsuffix]))                
+                elif cnt <= 0:
                     self._redis.hset("workers", self.worker_id, json.dumps([worker_start_time, "Wait Task", cond, cmdprefix, cmdsuffix]))
-                elif not self._condition_parse(cond):
-                    self._redis.hset("workers", self.worker_id, json.dumps([worker_start_time, "Wait Resource", cond, cmdprefix, cmdsuffix]))
                 else:
                     row_str = self._redis.rpop("pending")
                     if row_str is None:
@@ -173,7 +189,7 @@ class CAM(object):
                             tf = self._get_by_tid("running", self.running_job_id)
                             taskinfo = taskinfo if tf is None else json.dumps(tf) 
                             self._set_by_tid("running", self.running_job_id, taskinfo)
-                            if self._get_by_tid("running", self.running_job_id)[-1] == "KILLED":
+                            if self._get_by_tid("running", self.running_job_id)[-1].startswith("KILLED"):
                                 os.killpg(os.getpgid(self.p.pid), signal.SIGINT)
                                 #self.p.kill()
                                 log_warn("Task ", str(self.running_job_id), " has been killed.")
@@ -188,8 +204,8 @@ class CAM(object):
                         except:
                             pass
                     log_info("{0} Finished command: {1}".format(self.worker_id, cmd))
-                    log_info("Server: ", self._conf['server'], ":", str(self._conf['port']), ' v', self.__version__)
-                    self._remove_by_tid("running", self.running_job_id)
+                    self._store_finished_task()
+                    delattr(self, "server_disconnected")
             except:
                 log_warn("Server Disconnected.")
                 self.server_disconnected = True
@@ -214,35 +230,56 @@ class CAM(object):
         <br>`cam ls` will list both tasks and workers information.
         <br>`cam ls worker 30` will list all workers wile each column has at most 30 chars.
         <br>`cam ls task 30` will list all tasks wile each column has at most 30 chars.
+        <br>`cam ls finished` will list all finished tasks.
         """
         now = datetime.datetime.utcnow()
         log_info("Server: ", self._conf['server'], ":", str(self._conf['port']), ' v', self.__version__)
+        def get_lst(part):
+            return sorted([parse_json(d) for d in self._redis.lrange(part, 0, -1)], key=lambda x: -x[0])
         try:
             if type is None or type == "task":
-                pending = self._redis.lrange("pending", 0, -1)
-                running = self._redis.lrange("running", 0, -1)
-                res = pending + running
+                pending = get_lst("pending")
+                running = get_lst("running")
+                finished = get_lst("finished")[:3]
+                res = pending + running 
                 nres = []
                 for i in range(len(res)):
-                    data = parse_json(res[i])
+                    data = res[i]
                     st = datetime.datetime.fromisoformat(data[1])
-                    data[1] = str(now - st).split('.')[0]
+                    data[1] = time_diff(now, st)
                     if maxwidth is not None:
                         data[2] = data[2][:maxwidth]
                     nres.append(data)
+                nres += finished
                 print(table_list(nres, headers = ["ID", "Time", "Command", "Host/PID"]))
-                print()
+                print("Pending: ", len(pending), " Running: ", len(running), " Finished: ", len(finished))
             if type is None or type == "worker":
                 workers = self._redis.hgetall("workers")
                 info = []
                 for w in workers:
                     dt = parse_json(workers[w])
                     st = datetime.datetime.fromisoformat(dt[0])
-                    lst = [w, str(now - st).split('.')[0]] + dt[1:]
+                    lst = [w, time_diff(now, st)] + dt[1:]
                     if maxwidth is not None and len(lst) > 3:
                         lst[3] = lst[3][:maxwidth]
-                    info.append(lst)    
+                    info.append(lst)
+                info = sorted(info, key=lambda x: x[0])
                 print(table_list(info, headers = ["Worker/PID", "Up Time", "Status", "cond", "prefix", "suffix"]))
+                status = {"Wait Resource": 0, "Wait Task": 0, "Running": 0}
+                for ir in info:
+                    if ir[2].startswith("Running"):
+                        status["Running"] += 1
+                    else:
+                        status[ir[2]] += 1
+                for k in status:
+                    print("{0} : {1}; ".format(k, status[k]), end="")
+            if type == "finished":
+                nres = get_lst("finished")[::-1]
+                print(table_list(nres, headers = ["ID", "Time", "Command", "Host/PID", "FnishTime"]))
+                print("Total: ", len(nres))
+                
+
+                
         except:
             log_warn("Server Disconnected.")
 
@@ -262,7 +299,7 @@ class CAM(object):
             log_warn("The task will be removed: \n", self._remove_by_tid("pending", rid))
         rrow = self._get_by_tid("running", rid)
         if rrow is not None:
-            rrow[-1] = "KILLED"
+            rrow[-1] = "KILLED " + rrow[-1]
             self._set_by_tid("running", rid, json.dumps(rrow))
 
     def refresh(self, type = None):
